/*-
 * Copyright (C) @BABOLO  2008 May 23
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2008 May 23\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: getopt.c,v 1.1 2008/05/24 10:59:49 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sysexits.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include "parser.h"

#define NOPT  0
#define OPT0  1
#define OPT1  2
/* Multiparameters options are possible with OPTx where x > 1 */
#define ERROUT(A,B) {ex = (A); errno = (B); goto out;}

int
babolo_openopt(u_int32_t flags, babolo_opt **opt, int argc, char *const *argv, const char *ostr) {
    babolo_opt *next;
    int ex = EX_OK;
    u_char *c;
    int i;

    ifBLIN_QV4(flags)
        fprintf( stderr, "+babolo_openopt %08X %08X %d %08X %s\n"
               , flags, (u_int32_t)opt, argc, (u_int32_t)argv, ostr
               );
    if  (!opt) {
        ifBLIN_QV1(flags) warnx("babolo_openopt: no opt");
        ERROUT(EX_DATAERR, EFAULT);
    }
    if  (!argv) {
        ifBLIN_QV1(flags) warnx("babolo_openopt: no argv");
        ERROUT(EX_DATAERR, EFAULT);
    }
    next = *opt;
    if  (!((*opt) = malloc(sizeof(babolo_opt)))) {
        ifBLIN_QV1(flags) warnx("babolo_openopt: no mem #0");
        ERROUT(EX_OSERR, ENOMEM);
    }
    (*opt)->next = next;
    (*opt)->flags = flags & ~(Bpars_NOAL | Bpars_NPRM | Bpars_EOOP);
    (*opt)->argc = argc;
    (*opt)->argv = (u_char**)argv;
    (*opt)->ind = (flags & Bpars_NONU) ? 1 : 0;
    (*opt)->plc = "";
    (*opt)->arg = NULL;
    if  (!!ostr) {
        if  (!((*opt)->opts = calloc(1, 256))) {
            free(*opt);
            *opt = next;
            ifBLIN_QV1(flags) warnx("babolo_openopt: no mem #1");
            ERROUT(EX_OSERR, ENOMEM);
        }
        for (i = 0; ostr[i];) {
            c = &(*opt)->opts[((u_char*)ostr)[i++]];
            if  (ostr[i] == ':') {
                *c = OPT1;
                i++;
            } else {
                *c = OPT0;
        }   }
    } else if (!next) {
        free(*opt);
        *opt = next;
        ifBLIN_QV1(flags) warnx("babolo_openopt: ostr and next both NULLs");
        ERROUT(EX_DATAERR, EINVAL);
    } else {
        (*opt)->opts = next->opts;
        (*opt)->flags |= Bpars_NOAL;
    }
out:
    ifBLIN_QV4(flags) fprintf(stderr, "-babolo_openopt %d\n", ex);
    return(ex);
}

int
babolo_closeopt(babolo_opt **opt) {
    u_int32_t flags = BLIN_VER1;
    babolo_opt *next;
    int ex = 0;

    if  (!opt) {
        warnx("babolo_closeopt: no opt");
        ERROUT(EX_DATAERR, EFAULT);
    }
    if  (!!*opt) flags = (*opt)->flags;
    while (!!*opt) {
        ifBLIN_QV4(flags)
            fprintf(stderr, "+babolo_closeopt %08X %08X\n", (u_int32_t)opt, (u_int32_t)*opt);
        next = (*opt)->next;
        if  (!((*opt)->flags & Bpars_NOAL)) free((*opt)->opts);
        free(*opt);
        *opt = next;
    }
out:
    ifBLIN_QV4(flags) fprintf(stderr, "-babolo_closeopt %d\n", ex);
    return(ex);
}

int
babolo_getopt(babolo_opt **opt) {
    u_int32_t flags = BLIN_VER1;
    babolo_opt *next;
    int ex = 0;

    if  (!opt) {
        warnx("babolo_getopt: no opt");
        ERROUT(EX_DATAERR, EFAULT);
    }
    if  (!*opt) goto out;
    flags = (*opt)->flags;
in:
    ifBLIN_QV4(flags)
        fprintf( stderr, "+babolo_getopt %08X %08X ind=%d argc=%d\n"
               , (u_int32_t)opt, (u_int32_t)*opt, (*opt)->ind, (*opt)->argc
               );
    ex = 0;
    if  ((*opt)->flags & Bpars_EOOP) {
        goto out;
    } else if (*((*opt)->plc) == 0) {
        ifBLIN_QV5(flags)
            fprintf(stderr, " ++ %08X [%d]%s~\n", (u_int32_t)*opt, (*opt)->ind, (*opt)->plc);
        if  (!(*opt)->argv) {
            ifBLIN_QV1(flags) warnx("babolo_getopt: no argv");
            ERROUT(EX_DATAERR, EFAULT);
        }
        if  ((*opt)->ind >= (*opt)->argc) {
            /* Argument is absent */
            next = (*opt)->next;
            if  (!((*opt)->flags & Bpars_NOAL)) free((*opt)->opts);
            free(*opt);
            *opt = next;
            if  (!*opt) goto out;
            goto in;
        }
        if  (!((*opt)->plc = (*opt)->argv[(*opt)->ind++])) {
            ifBLIN_QV1(flags) warnx("babolo_getopt: no argv[%d]", (*opt)->ind);
            ERROUT(EX_DATAERR, EFAULT);
        }
        if  (*((*opt)->plc) != '-') {
            /* Argument is not an option */
            (*opt)->flags |= Bpars_EOOP;
            goto out;
        }
        ex = *(++(*opt)->plc);
        if  (ex == '-' && *(++(*opt)->plc) == 0) {
            /* "--" => end of options */
            ex = 0;
            goto out;
        }
        ++(*opt)->plc;
        if  (ex == 0) {
            /* Solitary '-', treat as a '-' option */
            ex = '-';
        }
    } else {
        ifBLIN_QV5(flags)
            fprintf(stderr, " @@ %08X [%d]%s~\n", (u_int32_t)*opt, (*opt)->ind, (*opt)->plc);
        ex = *((*opt)->plc++);
    }
    ifBLIN_QV5(flags) fprintf(stderr, " getopt %02X=%c\n", ex, ex);
    (*opt)->flags &= ~Bpars_NPRM;
    switch((*opt)->opts[ex]) {
    case NOPT:
        ifBLIN_QV5(flags)
            fprintf(stderr, " NOPT %08X [%d]%s~\n", (u_int32_t)*opt, (*opt)->ind, (*opt)->plc);
        if  (!((*opt)->flags & Bpars_ESYM)) {
            ifBLIN_QV1(flags) warnx("babolo_getopt: invalid option");
            ERROUT(-1, EINVAL);
        }
        break;
    case OPT0:
        ifBLIN_QV5(flags)
            fprintf(stderr, " OPT0 %08X [%d]%s~\n", (u_int32_t)*opt, (*opt)->ind, (*opt)->plc);
        break;
    default:
        setBpars_NPRM((*opt)->flags, 1);
mid:
        ifBLIN_QV5(flags)
            fprintf(stderr, " OPT1 %08X [%d]%s~\n", (u_int32_t)*opt, (*opt)->ind, (*opt)->plc);
        if  (*((*opt)->plc) != 0) {
            (*opt)->arg = (*opt)->plc;
        } else if ((*opt)->ind < (*opt)->argc) {
            (*opt)->arg = (*opt)->argv[(*opt)->ind++];
        } else {
            /* option-argument absent */
            u_int32_t f;

            next = (*opt)->next;
            if  (!((*opt)->flags & Bpars_NOAL)) free((*opt)->opts);
            f = (*opt)->flags;
            free(*opt);
            *opt = next;
            if  (!*opt) {
                if  (!(f & Bpars_EPRM)) {
                    ifBLIN_QV1(flags) warnx("babolo_getopt: no argument for -%c", ex);
                    ERROUT(-1, ENOENT);
                }
                goto out;
            }
            goto mid;
        }
        (*opt)->plc = "";
    }
    ifBLIN_QV5(flags)
        fprintf(stderr, " -- %08X [%d]%s~\n", (u_int32_t)*opt, (*opt)->ind, (*opt)->plc);
out:
    ifBLIN_QV4(flags) fprintf(stderr, "-babolo_getopt %d\n", ex);
    return(ex);
}

char *
babolo_getoptarg(babolo_opt **opt, int argn) {
    char *c;

    c = NULL;
    if  (!opt) {
        warnx("babolo_getoptarg: no opt");
        errno = EFAULT;
        goto out;
    }
    if  (!*opt) goto out;
    ifBLIN_QV4((*opt)->flags)
        fprintf(stderr, "+babolo_getoptarg %08X %08X\n", (u_int32_t)opt, (u_int32_t)*opt);
    if  (argn > 0) {
        ifBLIN_QV1((*opt)->flags) warnx("babolo_getoptarg: myltiargument flags not realised\n");
    } else if (((*opt)->flags & Bpars_NPRM) > argn) c = (*opt)->arg;
    ifBLIN_QV4((*opt)->flags) {
        if  (!!c) fprintf(stderr, "-babolo_getoptarg =%s~\n", c);
          else fprintf(stderr, "-babolo_getoptarg\n");
    }
out:
    return(c);
}

char *
babolo_getarg(babolo_opt **opt) {
    u_int32_t flags = BLIN_VER1;
    babolo_opt *next;
    char *c;

    c = NULL;
    if  (!opt) {
        warnx("babolo_getarg: no opt");
        errno = EFAULT;
        goto out;
    }
    if  (!*opt) goto out;
    flags = (*opt)->flags;
in:
    ifBLIN_QV4(flags)
        fprintf( stderr, "+babolo_getarg %08X %08X ind=%d argc=%d\n"
               , (u_int32_t)opt, (u_int32_t)*opt, (*opt)->ind, (*opt)->argc
               );
    if  (*((*opt)->plc) == 0) {
        ifBLIN_QV5(flags)
            fprintf(stderr, " ++ %08X [%d]%s~\n", (u_int32_t)*opt, (*opt)->ind, (*opt)->plc);
        if  (!(*opt)->argv) {
            ifBLIN_QV1((*opt)->flags) warnx("babolo_getarg: no argv");
            errno = EFAULT;
            goto out;
        }
        if  ((*opt)->ind >= (*opt)->argc) {
            /* Argument is absent */
            next = (*opt)->next;
            if  (!((*opt)->flags & Bpars_NOAL)) free((*opt)->opts);
            free(*opt);
            *opt = next;
            if  (!*opt) goto out;
            goto in;
        }
        if  (!((*opt)->plc = (*opt)->argv[(*opt)->ind++])) {
            ifBLIN_QV1(flags) warnx("babolo_getarg: no argv[%d]", (*opt)->ind);
            errno = EFAULT;
            goto out;
    }   }
    c = (*opt)->plc;
    (*opt)->plc = "";
    ifBLIN_QV5(flags)
        fprintf(stderr, " -- %08X [%d]%s~\n", (u_int32_t)*opt, (*opt)->ind, (*opt)->plc);
out:
    ifBLIN_QV4(flags) fprintf(stderr, "-babolo_getarg=%s~\n", c);
    return(c);
}
